from pathlib import Path

import click

from ProjectAliceSK.validate.src.ConfigValidation import ConfigValidation
from ProjectAliceSK.validate.src.DialogValidation import DialogValidation
from ProjectAliceSK.validate.src.InstallValidation import InstallValidation
from ProjectAliceSK.validate.src.TalkValidation import TalkValidation
from ProjectAliceSK.validate.src.Validation import Validation


class Validator:

	def __init__(self, skillPaths: list, verbosity: int):
		self._dirPath = Path(__file__).resolve().parent.parent
		self._skillPath = self._dirPath.parent.parent
		self._skillPaths = skillPaths
		self._verbosity = verbosity


	@staticmethod
	def indentPrint(indent: int, *args):
		click.echo(' ' * (indent - 1) + ' '.join(map(str, args)))


	def validate(self):
		err = 0
		dialog = DialogValidation()
		installer = InstallValidation()
		talk = TalkValidation()
		config = ConfigValidation()

		for skillPath in self._skillPaths:
			skill = Path(skillPath).resolve()
			if not skill.is_dir():
				self.indentPrint(0, click.style(f'The path {skillPath} is invalid', fg='red', bold=True))
				continue

			dialog.reset(skill)
			installer.reset(skill)
			talk.reset(skill)
			config.reset(skill)

			dialog.validate(self._verbosity)
			installer.validate()
			talk.validate()
			config.validate()

			if dialog.errorCode or installer.errorCode or talk.errorCode:
				err = 1
				self.indentPrint(0, click.style(f'{skill.name}', fg='red', bold=True), 'invalid')
				self.printErrors('Installer', installer)
				self.printErrors('Dialog files', dialog)
				self.printErrors('Talk files', talk)
				self.printErrors('Config file', config)
				self.indentPrint(0)
			else:
				self.indentPrint(0, click.style(f'{skill.name}', fg='green', bold=True), 'valid')

		return err


	def printErrors(self, name: str, validation: Validation):
		if validation.errorCode:
			self.indentPrint(2, click.style(f'{name}:', bold=True))
			click.echo(validation.errors)
		else:
			self.indentPrint(2, click.style(name, bold=True), 'valid')



import re
from collections import defaultdict
from typing import Match


class DialogTemplate:

	def __init__(self, dialogTemplate: dict, verbosity: int = 0):
		self._dialogTemplate = dialogTemplate
		self._verbosity = verbosity
		self._slots = dict()
		self._intents = dict()
		self._cleanedUtterances = dict()
		self._utteranceSlots = dict()

		self._initSlots()
		self._initIntents()
		self._initCleanedUtterances()
		self._initUtteranceSlots()


	def _initSlots(self):
		for slot in self._dialogTemplate.get('slotTypes', dict()):
			self._slots[slot['name']] = slot


	def _initIntents(self):
		for intent in self._dialogTemplate.get('intents', dict()):
			self._intents[intent['name']] = intent


	def _cleanUtterance(self, utterance: str) -> str:
		def upperRepl(match: Match) -> str:
			return match.group(1).upper()


		cleanUtterance = utterance.lower()
		if self._verbosity:
			cleanUtterance = re.sub(r'{.+?:=>(.+?)}', upperRepl, cleanUtterance)
		cleanUtterance = re.sub(r'[^a-zA-Z1-9 ]', '', cleanUtterance)
		return ' '.join(cleanUtterance.split())


	def _initCleanedUtterances(self):
		for intentName, intents in self.intents.items():
			self._cleanedUtterances[intentName] = defaultdict(list)
			for utterance in intents['utterances']:
				self._cleanedUtterances[intentName][self._cleanUtterance(utterance)].append(utterance)


	@staticmethod
	def _mapUtteranceSlots(utterance: str, slots: list) -> defaultdict:
		utteranceSlotMapping = defaultdict(list)
		slotNames = re.findall(r'{(.*?):=>(.*?)}', utterance)
		for slot in slots:
			for slotValue, slotName in slotNames:
				if slot['name'] == slotName:
					utteranceSlotMapping[slot['type']].append(slotValue)

		return utteranceSlotMapping


	def _initUtteranceSlots(self):
		for intentName, intents in self.intents.items():
			for utterance in intents['utterances']:
				self._utteranceSlots[intentName] = self._mapUtteranceSlots(utterance, intents.get('slots', []))


	@property
	def slots(self) -> dict:
		return self._slots


	@property
	def intents(self) -> dict:
		return self._intents


	@property
	def cleanedUtterances(self) -> dict:
		return self._cleanedUtterances


	@property
	def utteranceSlots(self) -> dict:
		return self._utteranceSlots



import click


class OptionEatAll(click.Option):
	"""
	taken from https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
	"""


	def __init__(self, *args, **kwargs):
		self.save_other_options = kwargs.pop('save_other_options', True)
		nargs = kwargs.pop('nargs', -1)
		assert nargs == -1, f'nargs, if set, must be -1 not {nargs}'
		super(OptionEatAll, self).__init__(*args, **kwargs)
		self._previous_parser_process = None
		self._eat_all_parser = None


	def add_to_parser(self, parser, ctx):

		def parser_process(value, state):
			# method to hook to the parser.process
			done = False
			value = [value]
			if self.save_other_options:
				# grab everything up to the next option
				while state.rargs and not done:
					for prefix in self._eat_all_parser.prefixes:
						if state.rargs[0].startswith(prefix):
							done = True
					if not done:
						value.append(state.rargs.pop(0))
			else:
				# grab everything remaining
				value += state.rargs
				state.rargs[:] = []
			value = tuple(value)

			# call the actual process
			self._previous_parser_process(value, state)


		retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
		for name in self.opts:
			# noinspection PyProtectedMember
			our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
			if our_parser:
				self._eat_all_parser = our_parser
				self._previous_parser_process = our_parser.process
				our_parser.process = parser_process
				break
		return retval
